from __future__ import annotations

import typing

import absorb
from .. import cli_outputs
from .. import cli_parsing

if typing.TYPE_CHECKING:
    from argparse import Namespace
    from typing import Any


def remove_command(args: Namespace) -> dict[str, Any]:
    if args.all:
        tracked_datasets = absorb.ops.get_tracked_tables()
    else:
        tracked_datasets = cli_parsing._parse_datasets(args)
    absorb.ops.stop_tracking_tables(tracked_datasets)
    cli_outputs._print_title('Stopped tracking')
    for dataset in tracked_datasets:
        cli_outputs._print_dataset_bullet(dataset)
    return {}



from __future__ import annotations

import typing

import absorb

if typing.TYPE_CHECKING:
    import polars as pl


class FourbyteDatatype(absorb.Table):
    source = 'fourbyte'
    write_range = 'append_only'
    range_format = 'id_range'

    # custom
    endpoint: str

    def get_schema(self) -> dict[str, type[pl.DataType] | pl.DataType]:
        import polars as pl

        return {
            'id': pl.Int64,
            'created_at': pl.Datetime,
            'text_signature': pl.String,
            'hex_signature': pl.String,
            'bytes_signature': pl.Binary,
        }

    def get_available_range(self) -> typing.Any:
        import requests

        data = requests.get(self.endpoint).json()
        max_id = max(result['id'] for result in data['results'])
        return (0, max_id)

    async def async_collect_chunk(
        self, data_range: typing.Any
    ) -> pl.DataFrame | None:
        return await async_scrape_4byte(
            url=self.endpoint, data_range=data_range
        )


class Functions(FourbyteDatatype):
    endpoint = 'https://www.4byte.directory/api/v1/signatures/'


class Events(FourbyteDatatype):
    endpoint = 'https://www.4byte.directory/api/v1/event-signatures/'


def get_tables() -> list[type[absorb.Table]]:
    return [Functions, Events]


async def async_scrape_4byte(
    url: str,
    data_range: tuple[int, int],
    wait_time: float = 0.1,
    min_id: int | None = None,
) -> pl.DataFrame:
    import aiohttp
    import polars as pl

    results = []
    async with aiohttp.ClientSession() as session:
        while True:
            # get page
            async with session.get(url) as response_object:
                response = await response_object.json()
                results.extend(response['results'])

            # scrape only until min_id is reached
            if min_id is not None:
                min_result_id = min(
                    result['id'] for result in response['results']
                )
                if min_result_id < min_id:
                    break

            # get next url
            url = response['next']
            if url is None:
                break

            # wait between responses
            if wait_time is not None:
                import asyncio

                await asyncio.sleep(wait_time)

    return pl.DataFrame(results, orient='row')



from __future__ import annotations

import typing

import absorb
from . import paths

if typing.TYPE_CHECKING:
    import typing_extensions


def get_default_config() -> absorb.Config:
    return {
        'version': absorb.__version__,
        'tracked_tables': [],
    }


def get_config() -> absorb.Config:
    import json

    try:
        with open(paths.get_config_path(), 'r') as f:
            config = json.load(f)
    except FileNotFoundError:
        return get_default_config()

    default_config = get_default_config()
    default_config.update(config)
    config = default_config

    if validate_config(config):
        return config
    else:
        raise Exception('invalid config format')


def write_config(config: absorb.Config) -> None:
    import json
    import os

    default_config = get_default_config()
    default_config.update(config)
    config = default_config

    if not validate_config(config):
        raise Exception('invalid config format')

    path = paths.get_config_path()
    os.makedirs(os.path.dirname(path), exist_ok=True)
    with open(path, 'w') as f:
        json.dump(config, f)


def validate_config(
    config: typing.Any,
) -> typing_extensions.TypeGuard[absorb.Config]:
    return (
        isinstance(config, dict)
        and {'tracked_tables'}.issubset(set(config.keys()))
        and isinstance(config['tracked_tables'], list)
    )


#
# # specific accessors
#


def get_tracked_tables() -> list[absorb.TrackedTable]:
    return get_config()['tracked_tables']


def start_tracking_tables(tables: list[absorb.TrackedTable]) -> None:
    import json

    config = get_config()
    tracked_tables = {
        json.dumps(table, sort_keys=True) for table in config['tracked_tables']
    }
    for table in tables:
        as_str = json.dumps(table, sort_keys=True)
        if as_str not in tracked_tables:
            config['tracked_tables'].append(table)
            tracked_tables.add(as_str)
    write_config(config)


def stop_tracking_tables(tables: list[absorb.TrackedTable]) -> None:
    import json

    tables_str = [json.dumps(table, sort_keys=True) for table in tables]

    config = get_config()
    config['tracked_tables'] = [
        table
        for table in config['tracked_tables']
        if json.dumps(table, sort_keys=True) not in tables_str
    ]

    write_config(config)



from __future__ import annotations

import typing
from . import table_paths

if typing.TYPE_CHECKING:
    T = typing.TypeVar('T')

    import datetime


class TableCoverage(table_paths.TablePaths):
    def get_available_range(self) -> typing.Any:
        return None

    def get_collected_range(self) -> typing.Any:
        import os
        import glob

        dir_path = self.get_dir_path()
        if not os.path.isdir(dir_path):
            return None

        glob_str = self.get_glob()
        if self.write_range == 'overwrite_all':
            files = sorted(glob.glob(glob_str))
            if len(files) == 0:
                return None
            elif len(files) == 1:
                return self.parse_file_path(files[0])['data_range']
            else:
                raise Exception('too many files')
        elif self.is_range_sortable():
            files = sorted(glob.glob(glob_str))
            start = self.parse_file_path(files[0])['data_range']
            end = self.parse_file_path(files[-1])['data_range']
            return [start, end]
        else:
            raise Exception()

    @classmethod
    def is_range_sortable(cls) -> bool:
        return cls.range_format is not None

    def get_min_collected_timestamp(self) -> datetime.datetime:
        raise NotImplementedError()

    def get_max_collected_timestamp(self) -> datetime.datetime:
        raise NotImplementedError()

    def get_min_available_timestamp(self) -> datetime.datetime:
        raise NotImplementedError()

    def get_max_available_timestamp(self) -> datetime.datetime:
        raise NotImplementedError()



from __future__ import annotations

import typing

import absorb

if typing.TYPE_CHECKING:
    import polars as pl


def scan(
    dataset: absorb.TableReference,
    *,
    parameters: dict[str, typing.Any] | None = None,
    scan_kwargs: dict[str, typing.Any] | None = None,
) -> pl.LazyFrame:
    import polars as pl

    table = absorb.ops.resolve_table(dataset, parameters=parameters)
    glob = table.get_glob()
    if scan_kwargs is None:
        scan_kwargs = {}
    try:
        return pl.scan_parquet(glob, **scan_kwargs)
    except Exception as e:
        if e.args[0].startswith('expected at least 1 source'):
            raise Exception('no data to load for ' + str(dataset))
        else:
            raise e


def load(dataset: absorb.TableReference, **kwargs: typing.Any) -> pl.DataFrame:
    """kwargs are passed to scan()"""
    import polars as pl

    try:
        return scan(dataset=dataset, **kwargs).collect()
    except pl.exceptions.ComputeError as e:
        if e.args[0].startswith('expected at least 1 source'):
            raise Exception('no data to load for ' + str(dataset))
        else:
            raise e


def write_file(*, df: pl.DataFrame, path: str) -> None:
    import os
    import shutil

    os.makedirs(os.path.dirname(path), exist_ok=True)

    tmp_path = path + '_tmp'
    if path.endswith('.parquet'):
        df.write_parquet(tmp_path)
    elif path.endswith('.csv'):
        df.write_csv(tmp_path)
    else:
        raise Exception('invalid file extension')
    shutil.move(tmp_path, path)



from __future__ import annotations

import typing

import absorb
from .. import cli_outputs
from .. import cli_parsing

if typing.TYPE_CHECKING:
    from argparse import Namespace
    from typing import Any


def add_command(args: Namespace) -> dict[str, Any]:
    import json
    import rich

    # parse inputs
    if args.all:
        track_datasets = absorb.ops.get_available_tables(exclude_parameters=True)
    else:
        track_datasets = cli_parsing._parse_datasets(args)

    # use snake case throughout
    for track_dataset in track_datasets:
        track_dataset['source_name'] = absorb.ops.names._camel_to_snake(
            track_dataset['source_name']
        )
        track_dataset['table_name'] = absorb.ops.names._camel_to_snake(
            track_dataset['table_name']
        )

    # filter already collected
    tracked = [
        json.dumps(table, sort_keys=True)
        for table in absorb.ops.get_tracked_tables()
    ]
    already_tracked = []
    not_tracked = []
    for ds in track_datasets:
        if json.dumps(ds, sort_keys=True) in tracked:
            already_tracked.append(ds)
        else:
            not_tracked.append(ds)
    track_datasets = not_tracked

    # check for invalid datasets
    sources = set(td['source_name'] for td in track_datasets)
    source_datasets = {
        source: [
            table.class_name() for table in absorb.ops.get_source_tables(source)
        ]
        for source in sources
    }
    for track_dataset in track_datasets:
        if (
            track_dataset['table_name']
            not in source_datasets[track_dataset['source_name']]
        ):
            raise Exception('invalid dataset:')

    # print dataset summary
    if len(already_tracked) > 0:
        cli_outputs._print_title('Already tracking')
        for dataset in already_tracked:
            cli_outputs._print_dataset_bullet(dataset)
        print()
    cli_outputs._print_title('Now tracking')
    if len(track_datasets) == 0:
        print('[no new datasets specified]')
    else:
        for dataset in track_datasets:
            cli_outputs._print_dataset_bullet(dataset)
        print()
        rich.print(
            'to proceed with data collection, use [white bold]absorb collect[/white bold]'
        )
    absorb.ops.start_tracking_tables(track_datasets)

    return {}



from __future__ import annotations

import typing

import absorb
from .. import cli_parsing

if typing.TYPE_CHECKING:
    from argparse import Namespace
    from typing import Any


def collect_command(args: Namespace) -> dict[str, Any]:
    if len(args.dataset) > 0:
        datasets = cli_parsing._parse_datasets(args)
    else:
        datasets = absorb.ops.config.get_tracked_tables()

    first = True
    for dataset in datasets:
        if not first:
            print()
        instance = absorb.Table.instantiate(dataset)
        if instance.write_range == 'append_only':
            data_ranges = cli_parsing._parse_ranges(
                args.range, range_format=instance.range_format
            )
        else:
            data_ranges = None
        instance.collect(
            data_range=data_ranges,
            dry=args.dry,
            overwrite=args.overwrite,
            verbose=args.verbose,
        )
        first = False

    return {}



from .catalog import *
from .collection import *
from .config import *
from .coverage import *
from .formatting import *
from .io import *
from .names import *
from .networking import *
from .paths import *
from .ranges import *



from __future__ import annotations

import typing

import absorb

if typing.TYPE_CHECKING:
    import polars as pl
    import flipside  # type: ignore


class Query(absorb.Table):
    source = 'flipside'
    write_range = 'overwrite_all'
    range_format = 'per_hour'
    parameters = {'name': str, 'sql': str}

    def get_schema(self) -> dict[str, type[pl.DataType] | pl.DataType]:
        raise NotImplementedError('implement schema for Query')

    def collect_chunk(self, data_range: typing.Any) -> pl.DataFrame:
        sql = self.parameters['sql']
        return query(sql)


_clients = {
    'default': None,
}


def get_flipside_api_key() -> str:
    import os

    api_key = os.environ.get('FLIPSIDE_API_KEY')
    if api_key is None or api_key == '':
        raise Exception(
            'set FLIPSIDE_API_KEY in env var or use set_flipside_api_key() function'
        )
    else:
        return api_key


def set_flipside_api_key(api_key: str) -> None:
    import os

    os.environ['FLIPSIDE_API_KEY'] = api_key


def get_flipside_client(
    *, use_default: bool = True, api_key: str | None = None
) -> flipside.Flipside:
    import flipside

    if use_default and _clients['default'] is not None:
        return _clients['default']
    else:
        if api_key is None:
            api_key = get_flipside_api_key()
        client = flipside.Flipside(api_key, 'https://api-v2.flipsidecrypto.xyz')
        _clients['default'] = client
        return client


def query(sql: str) -> pl.DataFrame:
    import polars as pl

    client = get_flipside_client()
    result = client.query(sql)

    return pl.DataFrame(
        result.rows,
        schema=result.columns,
        orient='row',
        infer_schema_length=len(result.rows),
    )



import absorb

import pytest


@pytest.mark.parametrize('table', absorb.ops.get_table_classes())
def test_tables_have_attrs(table: type[absorb.Table]) -> None:
    for attr in [
        'source',
        'write_range',
    ]:
        assert hasattr(table, attr), (
            'missing attribute '
            + attr
            + ' for '
            + str(table.source)
            + '.'
            + str(table.__name__)
        )


@pytest.mark.parametrize('table', absorb.ops.get_table_classes())
def test_tables_implement_get_schema(table: type[absorb.Table]) -> None:
    assert table.get_schema != absorb.Table.get_schema, (
        'missing get_schema() for '
        + str(table.source)
        + '.'
        + str(table.__name__)
    )


@pytest.mark.parametrize('table', absorb.ops.get_table_classes())
def test_tables_implement_collect_chunk(table: type[absorb.Table]) -> None:
    assert table.collect_chunk != absorb.Table.collect_chunk, (
        'missing collect_chunk() for '
        + str(table.source)
        + '.'
        + str(table.__name__)
    )


@pytest.mark.parametrize('table', absorb.ops.get_table_classes())
def test_table_parameter_names_valid(table: type[absorb.Table]) -> None:
    for name in table.parameter_types.keys():
        assert name not in [
            'base_name',
        ]
        assert '__' not in name
        assert not name.startswith('_')
        assert not name.endswith('_')



from __future__ import annotations

import typing
import polars as pl
import absorb


class BaseQuery(absorb.Table):
    source = 'dune'
    name_template = {'custom': '{name}'}

    def get_schema(self) -> dict[str, type[pl.DataType] | pl.DataType]:
        import spice

        query = self.parameters['query']
        spice_kwargs = self.parameters['spice_kwargs']
        spice_kwargs['limit'] = 0
        return dict(spice.query(query, **spice_kwargs).schema)


class FullQuery(BaseQuery):
    """collect the full output of a query"""

    write_range = 'overwrite_all'
    parameter_types = {
        'name': str,
        'query': str,
        'spice_kwargs': dict[str, typing.Any],
    }

    def collect_chunk(self, data_range: typing.Any) -> pl.DataFrame:
        import spice

        query = self.parameters['query']
        spice_kwargs = self.parameters['spice_kwargs']
        return spice.query(
            query, poll=True, include_execution=False, **spice_kwargs
        )


class AppendOnlyQuery(BaseQuery):
    """collect the output of a query, time-partitioned"""

    write_range = 'append_only'
    range_format = 'named_range'
    parameter_types = {
        'name': str,
        'query': str,
        'spice_kwargs': dict[str, typing.Any],
        'range_parameters': list[str],
    }

    def collect_chunk(self, data_range: typing.Any) -> pl.DataFrame | None:
        import spice

        query = self.parameters['query']
        spice_kwargs = self.parameters['spice_kwargs']
        spice_kwargs.setdefault('parameters', {})
        self.parameters.update(data_range)
        return spice.query(
            query, poll=True, include_execution=False, **spice_kwargs
        )


class CexLabels(absorb.Table):
    source = 'dune'
    write_range = 'overwrite_all'

    def get_schema(self) -> dict[str, type[pl.DataType] | pl.DataType]:
        import polars as pl

        return {
            'address': pl.String,
            'cex_name': pl.String,
            'distinct_name': pl.String,
            'added_by': pl.String,
            'added_date': pl.Datetime('ms'),
            'ecosystem': pl.String,
        }

    def collect_chunk(self, data_range: typing.Any) -> pl.DataFrame | None:
        import spice

        evm_cex_query = 'https://dune.com/queries/3237025'
        solana_cex_query = 'https://dune.com/queries/5124188'
        evm_cexes = spice.query(evm_cex_query).with_columns(ecosystem='EVM')
        solana_cexes = (
            spice.query(solana_cex_query)
            .drop('blockchain')
            .with_columns(ecosystem=pl.lit('solana'))
        )
        return pl.concat([evm_cexes, solana_cexes])


def get_tables() -> list[type[absorb.Table]]:
    return [CexLabels]



from __future__ import annotations

import typing
import absorb

if typing.TYPE_CHECKING:
    import argparse


def help_command(args: argparse.Namespace) -> dict[str, typing.Any]:
    import toolstr

    table = absorb.ops.resolve_table(args.dataset)
    schema = table.get_schema()

    toolstr.print_text_box(
        'dataset = ' + args.dataset, style='green', text_style='bold white'
    )
    for attr in [
        'source',
        'write_range',
        'index_by',
        'cadence',
    ]:
        if hasattr(table, attr):
            value = getattr(table, attr)
        else:
            value = None
        absorb.ops.print_bullet(key=attr, value=value)
    print()
    toolstr.print('[green bold]parameters[/green bold]')
    if table.parameters is None or len(table.parameter_types) == 0:
        print('- [none]')
    else:
        for key, value in table.parameter_types.items():
            if key in table.default_parameters:
                default = (
                    ' \[default = ' + str(table.default_parameters[key]) + ']'
                )
            else:
                default = ''
            absorb.ops.print_bullet(key=key, value=str(value) + default)
    print()
    toolstr.print('[green bold]schema[/green bold]')
    for key, value in schema.items():
        absorb.ops.print_bullet(key=key, value=value)

    return {}



from __future__ import annotations

import typing


bullet_styles = {
    'key_style': 'white bold',
    'bullet_style': 'green',
    'colon_style': 'green',
}


def print_bullet(
    key: str | None, value: str | None, **kwargs: typing.Any
) -> None:
    import toolstr

    toolstr.print_bullet(key=key, value=value, **kwargs, **bullet_styles)


def format_range(data_range: typing.Any) -> str:
    import datetime

    if isinstance(data_range, list):
        date_strs = [
            '-' if dt is None else dt.strftime('%Y-%m-%d') for dt in data_range
        ]
        return '\[' + ', '.join(date_strs) + ']'
    elif isinstance(data_range, datetime.datetime):
        return data_range.strftime('%Y-%m-%d')
    else:
        return str(data_range)
